/*
 * psum/psum-mutex.c --
 *
 * Let T_n be the execution time of a task with n processors. The
 * speedup S_n in the range [0..n] is defined as S_n = T_1/T_n. The
 * efficiency E_n in the range [0..100] is defined as E_n = S_n / n.
 *
 * gcc -Wall -O2 -o psum-mutex psum-mutex.c -lpthread
 *
 * Some execution times measured on dragon (virtual machine with 12 cores):
 *
 * dragon (naive):			real time	speedup	  efficiency
 * time ./psum-mutex 0 31		   1.577s	
 * time ./psum-mutex 1 31		  59.866s	1.000	    100.0 %
 * time ./psum-mutex 2 31		 172.183s	0.348        17.4 % 
 * time ./psum-mutex 4 31		 273.856s	0.219	      5.5 %
 * time ./psum-mutex 8 31		 386.934s	0.155         1.9 %
 *
 * dragon (array):			real time	speedup	  efficiency
 * time ./psum-mutex 0 31		   1.577s	
 * time ./psum-mutex 1 31		   1.527s	1.000	    100.0 %
 * time ./psum-mutex 2 31		   0.767s	1.991        99.5 % 
 * time ./psum-mutex 4 31		   0.385s	3.966	     99.2 %
 * time ./psum-mutex 8 31		   0.196s	7.791        97.4 %
 */

#define _POSIX_C_SOURCE 2

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>

#define MAXTHREADS 64

static const char *progname = "psum-mutex";

/* Global shared variables */
static unsigned long gsum = 0;		    /* Global sum */
static unsigned long nelems_per_thread;	    /* Number of elements to sum */
static unsigned long psum[MAXTHREADS];	    /* Array of partial results */

pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

/*
 * Simple loop to count up numbers - this is basic programming 101
 * code, no thrills, no threads (but a stupid O(n) algorithm since
 * we know an O(1) formula).
 */

static void sum(unsigned long end)
{
    unsigned long start = 0;
    unsigned long i;

    for (i = start; i < end; i++) {
	gsum += i;
    }
}

/*
 * Simple naive thread routine to count up a range of numbers.
 */

static void*
sum_mutex_naive(void *vargp)
{
    unsigned long myid = *((long *) vargp);	    /* Extract thread ID */
    unsigned long start = myid * nelems_per_thread; /* Start element index */
    unsigned long end = start + nelems_per_thread;  /* End element index */
    unsigned long i;

    for (i = start; i < end; i++) {
	(void) pthread_mutex_lock(&mutex);
	gsum += i;
	(void) pthread_mutex_unlock(&mutex);
    }
    return NULL;
}

/*
 * Smarter thread routine to count up a range on a non-shared global
 * variable; the main thread than sums up the partial results.
 */

static void*
sum_mutex_array(void *vargp)
{
    unsigned long myid = *((long *) vargp);	    /* Extract thread ID */
    unsigned long start = myid * nelems_per_thread; /* Start element index */
    unsigned long end = start + nelems_per_thread;  /* End element index */
    unsigned long i;

    for (i = start; i < end; i++) {
	psum[myid] += i;
    }
    return NULL;
}

/*
 * Smarter thread routine to count up a range on a non-shared local
 * register variable; the main thread than sums up the partial
 * results. Using a register variable may given better performance but
 * a smart compiler might do this optimization (holding data in a
 * register) anyway.
 */

static void*
sum_mutex_local(void *vargp)
{
    unsigned long myid = *((long *) vargp);	    /* Extract thread ID */
    unsigned long start = myid * nelems_per_thread; /* Start element index */
    unsigned long end = start + nelems_per_thread;  /* End element index */
    unsigned long i;
    register unsigned long sum = 0;

    for (i = start; i < end; i++) {
	sum += i;
    }
    psum[myid] = sum;
    return NULL;
}

int
main(int argc, char *argv[])
{
    const char *usage
	= "Usage: %s [-n numbers to sum up (logarithmic)] [-t threads] [-h] [-a naive|array|local]\n";
    int c;
    unsigned long i, nelems, log_nelems = 31, nthreads = 1,
	expected, myid[MAXTHREADS];
    pthread_t tid[MAXTHREADS];
    char *end;
    void* (*algorithm)(void*) = sum_mutex_naive;
    
    while ((c = getopt(argc, argv, "t:n:ha:")) >= 0) {
	switch (c) {
	case 't':
	    nthreads = strtoul(optarg, &end, 10);
	    if (*end) {
		fprintf(stderr, "%s: illegal number: %s\n", progname, optarg);
		exit(EXIT_FAILURE);
	    }
	    if (nthreads > MAXTHREADS) {
		fprintf(stderr, "%s: %lu exceeding thread limit %u\n",
			progname, nthreads, MAXTHREADS);
		exit(EXIT_FAILURE);
	    }
	    if (nthreads % 2) {
		fprintf(stderr,
			"%s: number of threads %lu must be a power of two\n",
			progname, nthreads);
		exit(EXIT_FAILURE);
	    }
	    break;
	case 'n':
	    log_nelems = strtoul(optarg, &end, 10);
	    if (*end) {
		fprintf(stderr, "%s: illegal number: %s\n", progname, optarg);
		exit(EXIT_FAILURE);
	    }
	    break;
	case 'a':
	    if (strcmp(optarg, "naive") == 0) {
		algorithm = sum_mutex_naive;
	    } else if (strcmp(optarg, "array") == 0) {
		algorithm = sum_mutex_array;
	    } else if (strcmp(optarg, "local") == 0) {
		algorithm = sum_mutex_local;
	    } else {
		fprintf(stderr, "%s: unknown algorithm '%s'\n",
			progname, optarg);
		exit(EXIT_FAILURE);
	    }
	    break;
	case 'h':
	    printf(usage, progname);
	    exit(EXIT_SUCCESS);
	case ':':
	case '?':
	    exit(EXIT_FAILURE);
	}
    }

    nelems = (1L << log_nelems);
    expected = (nelems * (nelems - 1)) / 2;

    if (nthreads == 0) {
	sum(nelems);
    } else if (nthreads > 0) {
	nelems_per_thread = nelems / nthreads;
	
	for (i = 0; i < nthreads; i++) {
	    myid[i] = i;
	    (void) pthread_create(&tid[i], NULL, algorithm, &myid[i]);
	}
	
	for (i = 0; i < nthreads; i++) {
	    (void) pthread_join(tid[i], NULL);
	}

	if (algorithm == sum_mutex_array
	    || algorithm == sum_mutex_local) {
	    for (i = 0; i < nthreads; i++) {
		gsum += psum[i];
	    }
	}
    }
	
    /*
     * Verify that we have produced the expected result...
     */

    if (gsum != expected) {
	fprintf(stderr, "%s: wrong result %lu (expected %lu)\n", progname,
		gsum, expected);
	return EXIT_FAILURE;
    }
    
    return EXIT_SUCCESS;
}
